"""
Algorithms for gene/transcript function prediction.

References
----------

Dozmorov, MG, Giles, CB, Wren, JD (2011). Predicting gene ontology from a
global meta-analysis of 1-color microarray experiments. BMC Bioinformatics, 12
Suppl 10:S14.

Wren, JD (2009). A global meta-analysis of microarray expression data to
predict unknown gene functions and estimate the literature-data divide.
Bioinformatics, 25, 13:1694-701.
"""
