"""
Signature analysis:

Problem
=======

Given a set of genes with numeric attributes (expression levels) and a
"database" of genes with numeric attributes, signature analysis provides
different methods of assessing the correlation between the two.

A concrete example: given a set of gene fold changes and a database of genes
that are positive and negative regulators of a certain pathway, is this pathway likely
to be up or down-regulated in this experiment?
"""


